import copy
import logging
import os.path
import re
import xml.etree.ElementTree as ET
from typing import Union

from remotemanager.connection.computers import BaseComputer, Substitution

logger = logging.getLogger(__name__)


class JUBETemplate(BaseComputer):
    """Extends BaseComputer to provide compatibility with JUBE platforms"""

    def __init__(self, template: str, platform: str):

        if os.path.exists(template):
            with open(template) as o:
                template = o.read()

        self._platform_path = ""
        self.platform_path = platform

        super().__init__(template=template)

    @classmethod
    def from_repo(
        cls,
        name: str,
        branch: str = "develop",
        repo: str = "https://gitlab.com/max-centre/JUBE4MaX",
        platform_name: str = "platform.xml",
        template_name: str = "submit.job",
        local_dir: Union[None, str] = None,
        **url_args,
    ):
        url = os.path.join("/", branch, "max-inputs/platforms/", name)
        url = f"{repo}/-/raw" + url

        if local_dir is None:
            local_dir = os.path.split(name.strip("/"))[-1]

            if local_dir == "" or local_dir is None:
                local_dir = os.getcwd()

        print(f"searching for {platform_name} & {template_name} at {url}")

        platform_path = os.path.join(local_dir, platform_name)
        cls.download_file(os.path.join(url, platform_name), platform_path)
        template_path = os.path.join(local_dir, template_name)
        cls.download_file(os.path.join(url, template_name), template_path)

        return cls(template=template_path, platform=platform_path)

    @property
    def platform_path(self) -> str:
        """Path to the platform file"""
        return self._platform_path

    @platform_path.setter
    def platform_path(self, path: str):
        if not os.path.exists(path):
            raise ValueError(f"Platform file {path} not found")

        self._platform_path = path

    def _extract_subs(self, link: bool = True) -> None:
        """
        Overrides BaseComputer sub extraction to instead parse from platform.xml
        """
        super()._extract_subs(link=False)
        logger.debug("generating subs specified in platform.xml")
        tree = ET.parse(self.platform_path)
        root = tree.getroot()
        # platform.xml specifies "substitutesets" which translate what's
        # in the template to the actual parameter name. Apply these changes here.
        empty = {}
        for subsituteset in root.findall("substituteset"):
            for change in subsituteset.iter():
                if change.tag != "sub":
                    logger.debug("skip row %s", change)
                    continue
                source = change.attrib["source"].strip("#").lower()
                dest = change.attrib["dest"].strip("$")

                if dest is None or dest == "":
                    # store this attribute for reporting if the destination is empty
                    empty[source] = change.attrib
                    continue

                if source in empty:
                    # if the source was previously flagged as missing a destination,
                    # but has one now, clear that flag here
                    del empty[source]

                # don't make unnecessary changes
                if source == dest:
                    continue

                if source in self.substitution_dict:

                    logger.debug("updating target %s to %s", source, dest)
                    obj = self.substitution_dict[source]

                    delattr(self, obj.name)

                    obj.target = dest
                    obj._name = dest

                    setattr(self, obj.name, obj)

        for attrib in empty.values():
            print(f"empty destination for {attrib}")

        cache = {}
        for parameterset in root.findall("parameterset"):
            for parameter in parameterset.iter():
                name = parameter.attrib["name"]
                mode = parameter.attrib.get("mode", None)

                default = parameter.text

                cache[name] = {"mode": mode, "default": default}

        for name, args in cache.items():
            default = args["default"]
            mode = args["mode"]

            if isinstance(default, str):
                for tmp in cache:
                    if f"${tmp}" in default:
                        mode = "python"

                default = default.strip()

            sub = Substitution(
                name=name, mode=mode, target=f"#{name}#", default=default
            )
            if hasattr(self, sub.name):
                delattr(self, sub.name)
            setattr(self, sub.name, sub)

    def reformat_template(self, platform: Union[None, str] = None) -> str:
        """
        converts the original JUBE template to one that Substitutions can interact with
        """

        if platform is not None:
            self.platform_path = platform

        template = copy.deepcopy(self.template)

        tree = ET.parse(self.platform_path)
        root = tree.getroot()
        for substituteset in root.findall("substituteset"):
            for substitution in substituteset.iter():
                if "source" in substitution.attrib:
                    source = substitution.attrib["source"]
                    dest = substitution.attrib.get("dest", None)

                    if dest is not None:
                        dest = f"#{dest.strip('$')}#"

                        template = template.replace(source, dest)

        return template

    def get_unevaluated_links(self, sub: Substitution) -> list:
        """
        Performs a regex search for unlinked JUBE $parameters within a value

        Args:
            sub: Substitution object to check
        """
        # only strings are valid targets
        if not isinstance(sub.value, str):
            return []
        # can lazily ignore any value not containing a $ char
        if "$" not in sub.value:
            return []
        # need to check for exact links using regex
        # this avoids bash $evaluations raising false positives
        cache = []
        for name in self.arguments:
            search = r"\$" + name + r"\b"
            if re.search(search, sub.value) is not None:
                cache.append(name)
        return cache

    def link_sub(self, sub: Substitution) -> None:
        """Performs recursive linking for target sub"""
        links = self.get_unevaluated_links(sub)

        # if the sub is fully linked, attempt an evaluation and return
        if len(links) == 0:
            logger.debug("%s has no children, evaluating and completing", sub.name)
            # set the temporary value, so these links do not permanently update values
            sub.temporary_value = attempt_eval(sub.value)
            return

        logger.debug("Performing recursive linking for sub %s", sub.name)
        # otherwise, we need to recursively continue
        value = sub.value  # store the value for treatment
        for name in links:
            logger.debug("\tevaluating child %s", name)
            obj = self.argument_dict[name]  # get the actual associated object
            # if this sub child object has children of itself, recurse the call
            sub_links = self.get_unevaluated_links(obj)  # search for any children
            if len(sub_links) != 0:
                logger.debug("\tlinked to: %s, recursing...", sub_links)
                self.link_sub(obj)
            # recursion ensures that this child has a valid value, collect it
            update = str(obj.value)
            # search and replace an exact match
            search = r"\$" + name + r"\b"
            # search for the exact string, continuing if not found
            # this prevents $tasks matching $taskspernode, for example
            if re.search(search, value) is None:
                continue
            value = re.sub(search, update, value)
        # set the temporary value
        sub.temporary_value = attempt_eval(value)

    def link_jube_subs(self) -> None:
        """Ensures proper linking of JUBE substitutions"""
        for sub in self.substitution_objects:
            self.link_sub(sub)

    def script(self, empty_treatment="local", **kwargs) -> str:
        """
        Performs a translation evaluation step
        before calling the BaseComputer script() method
        """
        for k, v in kwargs.items():
            if k in self.arguments:
                self.argument_dict[k].temporary_value = v

        # ensure substitutions are linked
        self.link_jube_subs()

        # update the template
        self._template = self.reformat_template()

        return super().script(empty_treatment=empty_treatment)


def attempt_eval(value: str):
    """Attempt to evaluate `value`"""
    try:
        return eval(value)
    except Exception:
        return value



import re
import time
from typing import Union

from remotemanager.storage.sendablemixin import SendableMixin
from remotemanager.utils import dir_delta

import os


class TrackedFile(SendableMixin):
    __slots__ = ["_remote_path", "_local_path", "_file", "_last_seen", "_size"]

    def __init__(self, local_path: str, remote_path: str, file: str):
        self._remote_path = remote_path
        self._local_path = local_path
        self._file = file

        self._last_seen = {"remote": -1, "local": -1}
        self._size = -1

    def __repr__(self) -> str:
        return self.local

    def __fspath__(self) -> str:
        return self.name

    @property
    def name(self) -> str:
        """Returns the filename"""
        return self._file

    @property
    def importstr(self) -> str:
        """
        Returns the filename without extension

        Suitable for python imports
        """
        return os.path.splitext(self._file)[0]

    @property
    def remote(self) -> str:
        """Returns the full remote path"""
        return os.path.join(self._remote_path, self.name)

    @property
    def local(self) -> str:
        """Returns the full local path"""
        return os.path.join(self._local_path, self.name)

    @property
    def remote_dir(self) -> str:
        """Returns the remote dir"""
        return self._remote_path

    @property
    def local_dir(self) -> str:
        """Returns the full local dir"""
        return self._local_path

    def relative_remote_path(self, other: str) -> str:
        """
        Return a path relative to `cwd`

        Args:
            other:
                working dir to compare against

        Returns:
            relative path
        """
        # if our remote path is an abspath, we already have what we need
        if os.path.isabs(self.remote_dir):
            return self.remote

        # we're already in the remote, just return the filename
        if self.remote_dir == other:
            return self.name

        # find the deepest shared path, treat it as a "root"
        stem = os.path.commonpath([self.remote_dir, other])
        # find how far down this stem is from `other`
        dirdelta = dir_delta(stem, other)
        # generate a ../ string that steps "down" to the common path
        down = "../" * dirdelta

        tmp_remote = self.remote_dir.replace(stem, "").strip("/")
        # rebuild up from our virtual root
        return os.path.join(down, tmp_remote, self.name)

    @property
    def content(self) -> Union[str, None]:
        """
        Attempts to read the file contents

        Returns None if the file cannot be read
        """
        if not os.path.isfile(self.local):
            return None
        with open(self.local, "r") as o:
            self.confirm_local()
            return o.read()

    def _write(
        self, content: Union[str, list], append: bool, add_newline: bool
    ) -> None:
        """
        Write to the file

        Args:
            content:
                Content to add
            append:
                Appends to file if True, overwrites otherwise
            add_newline:
                Finish the write with an extra newline if True
        """
        if not os.path.isdir(self.local_dir):
            os.makedirs(self.local_dir)
        # try to join lists, falling back on a basic str coercion
        if not isinstance(content, str):
            try:
                content = "\n".join(content)
            except TypeError:
                content = str(content)

        if append:
            mode = "a+"
        else:
            mode = "w+"

        with open(self.local, mode) as o:
            o.write(content)

            if add_newline and not content.endswith("\n"):
                o.write("\n")

        self.confirm_local()

    def write(self, content: Union[str, list], add_newline: bool = True) -> None:
        """
        Write `content` to the local copy of the file

        Args:
            content:
                content to write
            add_newline:
                enforces a newline character at the end of the write if True
                (default True)
        """
        self._write(content, append=False, add_newline=add_newline)

    def append(self, content: Union[str, list], add_newline: bool = True) -> None:
        """
        Append `content` to the local copy of the file

        Args:
            content:
                content to append
            add_newline:
                enforces a newline character at the end of the write if True
                (default True)
        """
        self._write(content, append=True, add_newline=add_newline)

    def confirm_local(self, t: Union[int, None] = None) -> None:
        """
        Confirm sighting of the file locally

        Args:
            t: Optionally set the time to `t` instead of time.time()
        """
        if t is None:
            t = int(time.time())
        self._last_seen["local"] = t

    def confirm_remote(self, t: Union[int, None] = None) -> None:
        """
        Confirm sighting of the file on the remote

        Args:
            t: Optionally set the time to `t` instead of time.time()
        """
        if t is None:
            t = int(time.time())
        self._last_seen["remote"] = t

    @property
    def exists_local(self) -> bool:
        """Returns True if the file exists locally"""
        return os.path.exists(self.local)

    def last_seen(self, where: str) -> int:
        """
        Returns the last_seen_<where>

        Where <where> is remote or local

        Args:
            where:
                remote or local
        """
        return self._last_seen[where]

    @property
    def last_seen_local(self) -> int:
        """Returns the time this file was last confirmed seen on the local machine"""
        return self.last_seen("local")

    @property
    def last_seen_remote(self) -> int:
        """Returns the time this file was last confirmed seen on the remote machine"""
        return self.last_seen("remote")

    @property
    def size(self) -> int:
        """
        Returns the filesize (needs to be set externally)

        -1 if not set
        """
        return self._size

    @size.setter
    def size(self, size: int) -> None:
        """Sets the filesize"""
        self._size = size

    def sub(self, source, target, mode: str = "python") -> bool:
        """
        Substitute source for target

        Args:
            source: file content to sub
            target: intended replacement
            mode: function to use, default to Python replace

        returns:
            True if a substitution was executed, False otherwise
        """
        content = self.content

        if content is None:
            return False

        mode = mode.lower()
        if mode == "python":
            content = self.content.replace(source, target)
        elif mode == "regex":
            content = re.sub(source, target, content)
        else:
            raise ValueError(f"mode {mode} not understood")

        self.write(content)
        return True



import importlib
import inspect
import logging
import os
import re
import typing

from remotemanager.storage.sendablemixin import SendableMixin
from remotemanager.utils.uuid import generate_uuid

logger = logging.getLogger(__name__)


_SCRIPT_TEMPLATE = """
{function}

if __name__ == '__main__':
\tkwargs = {args}

\tresult = {name}(**kwargs)
"""


class Function(SendableMixin):
    """
    Serialise function to an executable python file for transfer

    Args:
        func:
            python function for serialisation
    """

    def __init__(self, func: [typing.Callable, str], force_self: bool = False):
        logger.debug("creating new serialisable function for %s", func)

        self._uuid = ""
        # collect the source, name and signature
        if isinstance(func, str):
            source = func
            # need to extract the name
            name = "f"
            signature = ")"
            for line in func.split("\n"):
                if line.startswith("def"):
                    name, signature = line.lstrip("def ").split("(", maxsplit=1)

            signature = "(" + signature.strip(":")
        else:
            source = inspect.getsource(func)
            name = func.__name__
            signature = inspect.signature(func)

        if name in ["remote_load", "remote_dump"]:
            raise ValueError(f'function name "{name}" is protected')

        tmp = []
        indent = -1
        is_self = force_self  # default to False, allow forcing
        append = False  # don't begin the append until we hit the `def ...` line
        for line in source.split("\n"):
            if indent == -1 and line.strip().startswith("def"):
                if "(self" in line:
                    is_self = True
                append = True
                indent = line.index("def")

            if append:
                tmp.append(line[indent:])

        source = "\n".join(tmp).strip()

        signature = Function.prepare_signature(signature, is_self=is_self)

        self._signature = signature

        logger.debug("updated signature to %s", self._signature)
        rawsig = Function.get_raw_signature(source)
        logger.debug("raw typed signature detected as %s", self._signature)
        source = source.replace(rawsig, self._signature, 1)

        self._source = source
        self._fname = name

        # use 1:-1 to escape ()
        self._arglist = [a.strip() for a in self._signature[1:-1].split(",")]

        self._uuid = generate_uuid(self._source)

    def __call__(self, *args, **kwargs):
        return self.object(*args, **kwargs)

    @staticmethod
    def get_raw_signature(source):
        """
        Strips the signature as it is typed.
        inspect.signature does some formatting which can cause replacement to
        break in some conditions

        Args:
            source (str):
                raw source

        Returns:
            (str): raw signature as typed
        """
        definition = []
        for line in source.split("\n"):
            # we need to split from the function name for the raw signature
            if "(" in line:
                # split from first ( and up,
                # then join for any tuples defined as default
                line = "(" + "(".join(line.split("(")[1:])

            # appending the final ":" causes issues, strip that in a safe
            # manner then append
            definition.append(line.rstrip(" ").rstrip(":"))

            # we've reached the end of the definition
            if line.endswith(":"):
                break

        return "\n".join(definition)

    @staticmethod
    def prepare_signature(sig, is_self: bool = False) -> str:
        """
        Inserts *args and **kwargs into any signature that does not already
        have it

        Args:
            sig:
                inspect.signature(func)
            is_self:
                inspect.signature ignores `self`, yet we want to preserve it.
                Adds the argument if True

        Returns:
            (str): formatted sig
        """
        if isinstance(sig, str):
            sig = sig[: sig.rindex(")")]
            sig = sig.strip("(")
            args = [a.strip() for a in sig.split(",") if a.strip() != ""]
        else:
            args = [str(a) for a in sig.parameters.values()]

        if is_self and "self" not in args:
            args = ["self"] + args

        args_pattern = r"(\s|\()(\*{1}[\w]+)"
        kwargs_pattern = r"(\s|\()(\*{2}[\w]+)"

        has_args = re.search(args_pattern, str(sig)) is not None
        has_kwargs = re.search(kwargs_pattern, str(sig)) is not None

        if not has_args:
            if has_kwargs:
                args.insert(-1, "*args")
            else:
                args.append("*args")

        if not has_kwargs:
            args.append("**kwargs")

        return f'({", ".join(args)})'

    @property
    def name(self):
        """
        Function name
        """
        return self._fname

    @property
    def raw_source(self):
        """
        Function source
        """
        return self._source

    @property
    def source(self):
        """
        Function source

        Returns:
            (str): source code
        """
        return self._source

    @property
    def signature(self):
        return self._signature

    @property
    def args(self):
        return self._arglist

    @property
    def uuid(self):
        """
        Function uuid (64 characters)
        """
        return self._uuid

    @property
    def object(self):
        """
        Recreates the function object by writing out the source, and importing.

        Returns:
            typing.Callable:
                the originally passed function
        """

        tmp_file = ""
        try:
            tmp_file = os.path.abspath(f"{self.uuid}.py")

            with open(tmp_file, "w+") as o:
                o.write(self.source)

            spec = importlib.util.spec_from_file_location(self.uuid, tmp_file)
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)

            func_object = getattr(module, f"{self.name}")

        finally:
            os.remove(tmp_file)

        return func_object

    def dump_to_string(self, args):
        """
        Dump this function to a serialised string, ready to be written to a
        python file

        Args:
            args (dict):
                arguments to be used for this dumped run

        Returns:
            (str):
                serialised file
        """

        if args is None:
            args = {}

        return _SCRIPT_TEMPLATE.format(
            **{"function": self.source, "name": self.name, "args": args}
        )



"""Module to store the tokenizer class"""

import tokenize
from typing import Union


class Tokenizer:
    """
    generate_tokens() requires a file-like object to function

    This class mimics this behaviour by raising StopIteration once the
    end of the content is reached.
    """

    __slots__ = ["content", "row", "_tokens", "_names", "_numbers"]

    def __init__(self, content: str):
        self.content = content.split("\n")
        self.row = 0

        self._names = None
        self._numbers = None
        self._tokens = []
        self.tokenize()

    def __call__(self) -> Union[str, StopIteration]:
        if self.row == len(self.content):
            self.row = 0  # reset
            raise StopIteration
        line = self.content[self.row]
        self.row += 1
        return line

    def tokenize(self) -> list:
        """Runs the tokenization"""
        for token in tokenize.generate_tokens(self):
            self._tokens.append([token.type, token.string.strip()])
        return self.tokens

    @property
    def tokens(self) -> list:
        """
        Returns all stored tokens

        list format is [(type, string), ...]
        """
        return self._tokens

    @property
    def names(self) -> list:
        """Returns the derived name list"""
        if self._names is None:
            tmp = []
            for token in self.tokens:
                if token[0] == 1 and token[1] not in tmp:
                    tmp.append(token[1])
            self._names = tmp
        return self._names

    @property
    def numbers(self) -> list:
        """Returns derived number list"""
        if self._numbers is None:
            tmp = []
            for token in self.tokens:
                if token[0] == 2 and token[1] not in tmp:
                    tmp.append(token[1])
            self._numbers = tmp
        return self._numbers

    def exchange_name(self, a: str, b: str):
        """
        Exchanges name a with name b
        """
        for token in self.tokens:
            if token[1] == a:
                token[1] = b
        self._names = None  # invalidate the name cache

    @property
    def source(self):
        """
        Regenerate the source from the stored tokens

        untokenize is... unreliable, since it focuses on the
        repeatability of the round trip

        So we should reconstruct manually

        Key tokens:
        0: End marker
        1: Name
        2: Number
        3: String
        4: Newline
        5: Indent
        6: Dedent

        Returns:
            (str): reconstructed source
        """
        indent = 0
        output = []
        tmp = []
        for i, (ttype, token) in enumerate(self.tokens):
            if ttype in (0, 4):
                output.append("    " * indent + "".join(tmp))
                tmp = []
                continue
            if ttype == 5:
                output.append("    " * indent + "".join(tmp))
                tmp = []
                indent += 1
                continue
            if ttype == 6:
                output.append("    " * indent + "".join(tmp))
                tmp = []
                indent -= 1

            # If the current token is a registered name, and is followed
            # immediately by another, add a space to separate them
            try:
                nexttype = self.tokens[i + 1][0]
            except AttributeError:
                nexttype = 0

            if token in self.names and nexttype == 1:
                token += " "

            tmp.append(token)

        return "\n".join(output).strip()



import json
import os
import re
from datetime import datetime

date_format = "%d/%m/%Y %H:%M:%S"
manifest_filename = "{manifest_filename}"


class Manifest:

    def __init__(
        self, instance_uuid, path_override: str = None, name_override: str = None
    ):
        self.instance_uuid = instance_uuid

        self.filename = name_override or manifest_filename

        path = path_override or __file__
        if "." in path:
            path = os.path.split(path)[0]
        self.basepath = path

        self.runner_mode = False  # set True by runners, disables printing

    def now(self) -> str:
        return datetime.strftime(datetime.now(), date_format)

    def to_timestamp(self, timestring: str) -> int:
        return int(datetime.strptime(timestring, date_format).timestamp())

    @property
    def path(self):
        return os.path.join(self.basepath, self.filename)

    def get_data(self, path: str = None) -> list:
        if path is None:
            path = self.path
        if not os.path.exists(path):
            return []
        with open(path) as o:
            data = o.readlines()
        return data

    def write(self, string: str):
        string = f"{self.now()} {self.instance_uuid} {string.strip()}\n"
        with open(self.path, "a+") as o:
            o.write(string)

    def get(self, uuid: str) -> list:

        full_log = self.parse_log()
        uuid_select = full_log.get(uuid, [])
        if not uuid_select:
            return []

        log = [
            f"{datetime.fromtimestamp(line[0]).strftime(date_format)} {uuid} {line[1]}"
            for line in uuid_select
        ]
        return log

    def parse_log(self, path: str = None, string: str = None) -> dict:
        """Convert the log into a dict of {uuid: [time, log]}"""
        if string is not None:
            data = string.split("\n")
        else:
            data = self.get_data(path)

        date_regex = re.compile(r"\d{2}/\d{2}/\d{4}")

        cache = []
        output = {}

        uuid = None
        ts = 0
        for line in data:
            if re.match(date_regex, line):
                if len(cache) != 0:
                    data = [ts, "\n".join(cache)]
                    cache = []
                    try:
                        output[uuid].append(data)
                    except KeyError:
                        output[uuid] = [data]

                datestring, timestring, uuid, *content = line.split(" ")
                ts = self.to_timestamp(f"{datestring} {timestring}")

                cache.append("".join(content))
            else:
                cache.append(line)

        if len(cache) != 0:
            data = [ts, "".join(cache)]
            try:
                output[uuid].append(data)
            except KeyError:
                output[uuid] = [data]

        return output

    def last_time(self, state: str) -> dict:
        """Takes parsed log, returning the last time state appeared for each uuid"""
        data = self.parse_log()
        output = {}
        for uuid, log in data.items():
            if uuid not in output:
                output[uuid] = None
            for line in log:
                if line[1].strip().lower() == state.lower():
                    output[uuid] = line[0]

        if not self.runner_mode:
            print(json.dumps(output))
        return output


# DATASET_CONTENT #
